import numpy as np

from . import assert_array_equal, get_var_loader

load_var = get_var_loader('test_integers.mat')

neg_numbers = [[0, -1, -2, -3, -4, -5, -6, -7]]
pos_numbers = [[0, 1, 2, 3, 4, 5, 6, 7]]


def test_i8():
    i8 = load_var('i8')
    assert_array_equal(i8, np.array(neg_numbers, dtype='int8'))


def test_i16():
    i16 = load_var('i16')
    assert_array_equal(i16, np.array(neg_numbers, dtype='int16'))


def test_i32():
    i32 = load_var('i32')
    assert_array_equal(i32, np.array(neg_numbers, dtype='int32'))


def test_i64():
    i64 = load_var('i64')
    assert_array_equal(i64, np.array(neg_numbers, dtype='int64'))


def test_u8():
    u8 = load_var('u8')
    assert_array_equal(u8, np.array(pos_numbers, dtype='uint8'))


def test_u16():
    u16 = load_var('u16')
    assert_array_equal(u16, np.array(pos_numbers, dtype='uint16'))


def test_u32():
    u32 = load_var('u32')
    assert_array_equal(u32, np.array(pos_numbers, dtype='uint32'))


def test_u64():
    u64 = load_var('u64')
    assert_array_equal(u64, np.array(pos_numbers, dtype='uint64'))


def test_logical():
    b = load_var('b')
    assert_array_equal(b, np.array([[0, 1, 0, 0, 0, 1, 1]], dtype='bool8'))



import numpy as np

from . import assert_array_equal, get_matfile

matfile = get_matfile('test_arrays.mat')


def test_scalar():
    a0 = matfile.load_variable('a0')
    assert_array_equal(a0, np.array([[1.]]))


def test_1d():
    a1 = matfile.load_variable('a1')
    assert_array_equal(a1, np.array([[1., 2., 3.]]))


def test_2d():
    a2 = matfile.load_variable('a2')
    assert_array_equal(a2, np.array([[1., 2., 3.], [4., 5., 6.]]))


def test_3d():
    a3 = matfile.load_variable('a3')
    assert_array_equal(a3, np.arange(1., 24 + 1).reshape(2, 3, 4, order='F'))


def test_index_scalar():
    a0 = matfile['a0']
    value = a0[0, 0]
    assert value == 1.


def test_index_column():
    a2 = matfile['a2']
    assert_array_equal(a2[:, 0], np.array([1., 4.]))


def test_index_3d():
    a3 = matfile['a3']
    assert_array_equal(
        a3[:, :, 0],
        np.array([
            [1., 3., 5.],
            [2., 4., 6.],
        ]),
    )
    assert_array_equal(
        a3[:, :, 1],
        np.array([
            [7., 9., 11.],
            [8., 10., 12.],
        ]),
    )
    assert_array_equal(
        a3[:, :, 2],
        np.array([
            [13., 15., 17.],
            [14., 16., 18.],
        ]),
    )
    assert_array_equal(
        a3[:, :, 3],
        np.array([
            [19., 21., 23.],
            [20., 22., 24.],
        ]),
    )



import collections.abc
import importlib.resources
import os
import pathlib
from typing import Any, Dict, Generator, List, Type, Union

import h5py
import numpy as np

from .errors import MatlabDecodeError
from .loaders import (AbstractLoader, CellLoader, CharLoader, H5Object,
                      LogicalLoader, NumericLoader, StructLoader)

__version__ = importlib.resources.read_text(__name__, '__version__')

__all__ = [
    'Hdf5Matfile',
    'load_hdf5mat',
]


class Hdf5Matfile(collections.abc.Mapping):
    """Load data from an v7.3 *.mat file. Only reading is supported, no writing.

    Usage
    -----

    To load all the variables from the file, use |load_file|:

    .. code-block:: python

        with Hdf5Matfile(filename) as file:
            data = file.load_file()

    To fully load a specific variable from disk, use |load_variable|:

    .. code-block:: python

        with Hdf5Matfile(filename) as file:
            results = file.load_variable('results')

    For partial loading, a mapping/dict-like interface is also supported:

    .. code-block:: python

        with Hdf5Matfile(filename) as file:
            results = file['results']
            time = results[0, :]
            disp = results[1, :]
            ...

    If you're not using a context manager, make sure to close the file after
    you're done:

    .. code-block:: python

        file = Hdf5Matfile(filename)
        data = file.load_file()
        ...
        file.close()

    By default, arrays are not squeezed; since MATLAB represents even scalars
    as 2-D arrays, this means that something you expect to be a scalar will in
    fact be a 1-by-1 np.ndarray. You can change this by passing ``squeeze=True``
    to the constructor:

    .. code-block:: python

        with Hdf5Matfile(filename, squeeze=True) as file:
            data = file.load_file()

    Supported data types
    --------------------

    Data type support is pretty limited; this isn't a terribly fancy class.
    Supported MATLAB data types, and the Python objects or NumPy dtypes they map
    to:

    ===============  =============  =============
    MATLAB type      Python object  NumPy dtype
    ===============  =============  =============
    cell             np.ndarray     object
    char             str            n/a
    double           np.ndarray     double
    int8             np.ndarray     int8
    int16            np.ndarray     int16
    int32            np.ndarray     int32
    int64            np.ndarray     int64
    logical          np.ndarray     bool8
    single           np.ndarray     single
    struct (scalar)  dict           n/a
    struct (array)   np.ndarray     object (dict)
    uint8            np.ndarray     uint8
    uint16           np.ndarray     uint16
    uint32           np.ndarray     uint32
    uint64           np.ndarray     uint64
    ===============  =============  =============

    .. |load_file| :method:`Hdf5Matfile.load_file`
    .. |load_variable| :method:`Hdf5Matfile.load_variable`
    """
    def __init__(self, filename: os.PathLike, squeeze: bool = False):
        """Open a MATLAB v7.3 *.mat file.

        Parameters
        ----------
        filename : path_like
            Path to the *.mat file.

        squeeze : bool, int, optional
            If True, squeeze loaded arrays (remove dimensions with size 1). If
            the array only has one element, it is extracted from the array
            completely (instead of returning a 0-d array).
        """
        filepath = pathlib.Path(filename).resolve()
        try:
            self._h5file = h5py.File(filepath, 'r')
        except OSError as e:
            # Why are all the h5py errors just 'OSError'... ugh
            if 'No such file or directory' in str(e):
                raise FileNotFoundError(f'{filepath}') from e
            elif 'file signature not found' in str(e):
                raise OSError(f'Could not open {filepath} as HDF5 file') from e
            else:
                raise e
        self.filepath = filepath
        self.squeeze = squeeze

    #=============================================
    # Context manager stuff
    #=============================================
    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self):
        """Close the underlying h5py.File object."""
        self._h5file.close()

    @property
    def closed(self):
        try:
            self._h5file.id.fileno
        except TypeError:
            return True
        return False

    #=============================================
    # Public interface
    #=============================================
    def load_file(self):
        """Load the entire file.

        Returns
        -------
        dict
            Dict whose keys are the top-level variable names.
        """
        d = {}
        for varname, h5object in self._h5file.items():
            if varname.startswith('#'):
                continue
            d[varname] = self.get_loader(h5object)[()]
        return d

    def load_variable(self, varname: str):
        """Load a specific variable from the file.

        Parameters
        ----------
        varname : str
            The name of the variable to load.
        """
        h5object = self.get_h5object(varname)
        return self.get_loader(h5object)[()]

    #=============================================
    # Mapping interface
    #=============================================
    def __getitem__(self, varname: str):
        h5object = self.get_h5object(varname)
        return self.get_loader(h5object)

    def __iter__(self) -> Generator[str, Any, Any]:
        for key in self._h5file.keys():
            if key.startswith('#'):
                continue
            yield key

    def __len__(self):
        return len(list(iter(self)))

    #=============================================
    # Loaders
    #=============================================
    _loader_dispatch: Dict[str, Type[AbstractLoader]] = {}

    @classmethod
    def register_loader(cls, matlab_class: str, loader: Type[AbstractLoader]):
        cls._loader_dispatch[matlab_class] = loader

    def get_loader(self, h5object: Union[H5Object, h5py.Reference]):
        """Get the Loader object for a given high-level h5py object.

        Reference objects are de-referenced, and a loader for the underlying
        data is returned.
        """
        if isinstance(h5object, h5py.Reference):
            return self.get_loader(self._h5file[h5object])

        try:
            matlab_class = h5object.attrs['MATLAB_class'].decode()
        except KeyError:
            raise MatlabDecodeError('item does not have a MATLAB_class '
                                    'attribute and cannot be decoded')

        try:
            loader = self._loader_dispatch[matlab_class]
        except KeyError as e:
            raise MatlabDecodeError('Unsupported MATLAB class'
                                    f' {matlab_class!r}') from e

        return loader(h5object, parent=self)

    def get_h5object(self, varname: str):
        """Get the h5py high-level object for varname."""
        if varname.startswith('#'):
            raise KeyError(f'{varname!r} is not a MATLAB variable.')

        return self._h5file[varname]

    def _process(self, item):
        if isinstance(item, np.ndarray):
            # MATLAB arrays are column-major
            item = item.transpose()
            if self.squeeze:
                item = self._squeeze(item)

        return item

    @staticmethod
    def _squeeze(a: np.ndarray):
        if a.size == 1:
            squeezed = a.flat[0]
        else:
            squeezed = np.squeeze(a)
        return squeezed


# Collections
Hdf5Matfile.register_loader('struct', StructLoader)
Hdf5Matfile.register_loader('cell', CellLoader)

# Floats
Hdf5Matfile.register_loader('single', NumericLoader)
Hdf5Matfile.register_loader('double', NumericLoader)

# Integers
Hdf5Matfile.register_loader('logical', LogicalLoader)
Hdf5Matfile.register_loader('int8', NumericLoader)
Hdf5Matfile.register_loader('int16', NumericLoader)
Hdf5Matfile.register_loader('int32', NumericLoader)
Hdf5Matfile.register_loader('int64', NumericLoader)
Hdf5Matfile.register_loader('uint8', NumericLoader)
Hdf5Matfile.register_loader('uint16', NumericLoader)
Hdf5Matfile.register_loader('uint32', NumericLoader)
Hdf5Matfile.register_loader('uint64', NumericLoader)

# Strings
Hdf5Matfile.register_loader('char', CharLoader)


def load_hdf5mat(filename: os.PathLike,
                 variables: Union[str, List[str]] = None,
                 squeeze: bool = False):
    """Load a MATLAB v7.3 *.mat file. See |Hdf5Matfile| for limitations and
    supported data types.

    Parameters
    ----------
    filename : path_like
        Path to the *.mat file.

    variables : str, list[str], optional
        Variable name(s) to load. Default is None, which loads the entire file.

    squeeze : bool, optional
        If True, squeeze arrays and pop scalars. (default: False)


    .. |Hdf5Matfile| :class:`Hdf5Matfile`
    """
    with Hdf5Matfile(filename, squeeze=squeeze) as file:
        if variables is None:
            data = file.load_file()
        elif isinstance(variables, str):
            data = file.load_variable(variables)
        else:
            data = {var: file.load_variable(var) for var in variables}

    return data



from __future__ import annotations

import abc
from typing import Any, Union

import h5py
import numpy as np

__all__ = [
    'AbstractLoader',
    'CellLoader',
    'CharLoader',
    'LogicalLoader',
    'NumericLoader',
    'StructLoader',
]


def row_major_index(index):
    if isinstance(index, tuple):
        return index[::-1]
    else:
        return index


H5Object = Union[h5py.Dataset, h5py.Group]


class AbstractLoader(abc.ABC):
    def __init__(self, h5object: H5Object, parent):
        self.h5object = h5object
        self.parent = parent

    #=============================================
    # Representation and metadata
    #=============================================
    def __repr__(self) -> str:
        return f'<MATLAB {self.matlab_class} "{self.name}": shape {self.shape}>'

    @property
    def matlab_class(self):
        return self.h5object.attrs['MATLAB_class'].decode()

    @property
    def name(self):
        return self.h5object.name.split('/')[-1]

    @property
    @abc.abstractmethod
    def shape(self):
        ...

    def is_empty(self):
        return 'MATLAB_empty' in self.h5object.attrs

    #=============================================
    # Loading interface
    #=============================================
    def __getitem__(self, index):
        return self.load(index)

    @abc.abstractmethod
    def load(self, index) -> Any:
        """Load the specified item from disk."""
        ...


class StructLoader(AbstractLoader):
    """Loader for MATLAB type `struct`. Returns arrays of dict."""
    h5object: h5py.Group

    @property
    def shape(self):
        if self._is_struct_array():
            return self._get_struct_array_shape()
        else:
            return (1, 1)

    def _get_struct_array_shape(self):
        pointers = dict(self.h5object)
        _, refarray = pointers.popitem()
        return refarray.shape[::-1]

    def load(self, index) -> np.ndarray:
        if self._is_struct_array():
            s = self._load_array(index)
        else:
            s = self._load_scalar(index)
        return self.parent._process(s)

    def _load_scalar(self, index):
        d = {
            fieldname: self.parent.get_loader(item)[()]
            for fieldname, item in self.h5object.items()
        }
        return np.array([[d]], dtype='O')[index]

    def _load_array(self, index):
        c_index = row_major_index(index)

        # Get an item from the struct to figure out how big it is, then stick it
        # back in the dict. I have no idea if there's a cleaner way, I just need
        # to inspect a single item *before* looping!
        pointers = dict(self.h5object)
        fieldname, refarray = pointers.popitem()
        pointers[fieldname] = refarray

        # Figure out how much we're pulling out by indexing the sample refarray
        sample_refs = refarray[c_index]

        if isinstance(sample_refs, h5py.Reference):
            # Indexing pulled out a single item from the array, just return a
            # dict
            return {
                fieldname: self.parent.get_loader(refarray[c_index])[()]
                for fieldname, refarray in pointers.items()
            }

        # Initialize array of dict
        a = np.empty(sample_refs.shape, dtype='O')
        for i, _ in enumerate(a.flat):
            a[i] = dict()

        for fieldname, refarray in pointers.items():
            for i, ref in enumerate(refarray[c_index].flat):
                a.flat[i][fieldname] = self.parent.get_loader(ref)[()]

        return a

    def _is_struct_array(self):
        """Determine whether the given MATLAB struct is scalar or not."""
        for field in self.h5object.values():
            # MATLAB represents scalar structs and struct arrays differently
            # within HDF5. Scalar structs are ordinary groups with named
            # datasets and/or subgroups. Struct arrays, however, are represented
            # by a group with arrays of references. The arrays all have the same
            # size (that of the struct array itself), and are grouped by field
            # name.
            #
            # If the fields in the struct are *not* assigned a MATLAB_class,
            # then they're not actual objects. This is what differentiates a
            # struct array from a cell array -- the cell array is assigned a
            # MATLAB_class, and the fields of a struct array are not.
            try:
                matlab_class = field.attrs['MATLAB_class']
            except KeyError:
                isarray = True
                break
        else:
            # Executes if break doesn't fire
            isarray = False

        return isarray


class DatasetLoader(AbstractLoader):
    h5object: h5py.Dataset

    @property
    def shape(self):
        if self.is_empty():
            return ()
        else:
            return self.h5object.shape[::-1]


class CellLoader(DatasetLoader):
    """Loader for the MATLAB type `cell`. Returns arrays with dtype 'object'."""
    def load(self, index) -> np.ndarray:
        cellrefs = self.h5object[row_major_index(index)]
        if isinstance(cellrefs, h5py.Reference):
            # Indexed a single item, completely extracting it from the array
            return self.parent.get_loader(cellrefs)[()]

        a = np.empty(cellrefs.shape, dtype='O')
        for i, ref in enumerate(cellrefs.flat):
            a.flat[i] = self.parent.get_loader(ref)[()]
        return self.parent._process(a)


class NumericLoader(DatasetLoader):
    """Loader for MATLAB numeric types."""
    def load(self, index) -> np.ndarray:
        if 'MATLAB_empty' in self.h5object.attrs:
            return np.array([], dtype=self.h5object.dtype)
        return self.parent._process(self.h5object[row_major_index(index)])


class LogicalLoader(NumericLoader):
    """Loader for MATLAB type `logical`."""
    def load(self, index) -> np.ndarray:
        return super().load(index).astype('bool8')


class CharLoader(DatasetLoader):
    """Loader for the MATLAB type `char`. Returns str.

    Multi-dimensional `char` arrays are flattened to 1-D and returned as str.
    """
    def load(self, index) -> str:
        if self.is_empty():
            return ''
        c_index = row_major_index(index)
        return self.h5object[c_index].tobytes('F').decode('utf-16')



import importlib.resources

import numpy as np

from hdf5matfile import Hdf5Matfile, load_hdf5mat


def resource_path(name):
    return importlib.resources.path(__name__, name)


def get_var_loader(resource):
    def load_var(var):
        with resource_path(resource) as p:
            return load_hdf5mat(p, var)

    return load_var


def get_matfile(resource, squeeze=False):
    with resource_path(resource) as p:
        return Hdf5Matfile(p, squeeze)


def assert_array_equal(a: np.ndarray, b: np.ndarray):
    """Assert that two arrays have the same values, dtype, and shape."""
    assert a.dtype == b.dtype
    np.testing.assert_array_equal(a, b)


def assert_array_close(actual: np.ndarray,
                       desired: np.ndarray,
                       rtol: float = 1e-7,
                       atol: float = 0):
    """Assert that two arrays have the same values (within tolerance),
    dtype, and shape.
    """
    assert desired.dtype == actual.dtype
    assert desired.shape == actual.shape
    np.testing.assert_allclose(actual, desired, rtol, atol)



import numpy as np
from . import assert_array_equal, get_var_loader

load_var = get_var_loader('test_cells.mat')


def test_1d():
    c1 = load_var('c1')
    assert c1.dtype == np.dtype('O')
    assert c1.shape == (1, 3)
    a1, a2, a3 = c1.flat
    assert_array_equal(a1, np.array([[1.0]], dtype='double'))
    assert_array_equal(a2, np.array([[2.0]], dtype='double'))
    assert_array_equal(a3, np.array([[3.0]], dtype='double'))


def test_1d_arrays():
    c2 = load_var('c2')
    assert c2.dtype == np.dtype('O')
    assert c2.shape == (1, 2)
    a1 = c2.flat[0]
    a2 = c2.flat[1]
    assert_array_equal(a1, np.array([[1., 2., 3., 4.]]))
    assert_array_equal(a2, np.array([[5., 6., 7., 8.]]))


def test_2d():
    c3 = load_var('c3')
    assert c3.dtype == np.dtype('O')
    assert c3.shape == (2, 2)
    a11 = c3[0, 0]
    a12 = c3[0, 1]
    a21 = c3[1, 0]
    a22 = c3[1, 1]
    assert_array_equal(a11, np.array([[1., 2., 3., 4.]]))
    assert_array_equal(a12, np.array([[5., 6., 7., 8.]]))
    assert_array_equal(a21, np.array([[4., 3., 2., 1.]]))
    assert_array_equal(a22, np.array([[8., 7., 6., 5.]]))



import numpy as np

from . import assert_array_equal, get_var_loader

load_var = get_var_loader('test_chars.mat')


def test_1d():
    ch1 = load_var('ch1')
    assert ch1 == 'hello world!'


def test_2d():
    ch2 = load_var('ch2')
    assert ch2 == 'helloworld'


def test_chars_in_cells():
    ch3 = load_var('ch3')
    assert_array_equal(ch3, np.array([['hello', 'world']], dtype=object))



import numpy as np

from . import assert_array_equal, get_var_loader

load_var = get_var_loader('test_structs.mat')


def test_simple():
    s1 = load_var('s1')
    assert s1.shape == (1, 1)

    d = s1[0, 0]
    assert_array_equal(d['a'], np.array([[3]], dtype='double'))
    assert d['b'] == 'foobar'


def test_non_scalar():
    s2 = load_var('s2')
    assert s2.shape == (1, 2)

    d1 = s2[0, 0]
    assert_array_equal(d1['a'], np.array([[1]], dtype='double'))
    assert d1['b'] == 'foo'

    d2 = s2[0, 1]
    assert_array_equal(d2['a'], np.array([[2]], dtype='double'))
    assert d2['b'] == 'bar'


def test_nested():
    s3 = load_var('s3')
    assert s3.shape == (1, 1)

    # s1
    s1 = s3[0, 0]['s1']
    assert s1.shape == (1, 1)

    d = s1[0, 0]
    assert_array_equal(d['a'], np.array([[3]], dtype='double'))
    assert d['b'] == 'foobar'

    # s2
    s2 = s3[0, 0]['s2']
    assert s2.shape == (1, 2)

    d1 = s2[0, 0]
    assert_array_equal(d1['a'], np.array([[1]], dtype='double'))
    assert d1['b'] == 'foo'

    d2 = s2[0, 1]
    assert_array_equal(d2['a'], np.array([[2]], dtype='double'))
    assert d2['b'] == 'bar'
